%% ``The contents of this file are subject to the Erlang Public License,
%% Version 1.1, (the "License"); you may not use this file except in
%% compliance with the License. You should have received a copy of the
%% Erlang Public License along with this software. If not, it can be
%% retrieved via the world wide web at http://www.erlang.org/.
%% 
%% Software distributed under the License is distributed on an "AS IS"
%% basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See
%% the License for the specific language governing rights and limitations
%% under the License.
%% 
%% The Initial Developer of the Original Code is Ericsson Utvecklings AB.
%% Portions created by Ericsson are Copyright 1999, Ericsson Utvecklings
%% AB. All Rights Reserved.''
%% 
%%     $Id$
%%

%%%----------------------------------------------------------------------
%%% File    : fprof_server.erl
%%% Author  : Raimo Niskanen <raimo@erix.ericsson.se>
%%% Purpose : The analyse server for fprof.
%%% Created : 18 Jun 2001 by Raimo Niskanen <raimo@erix.ericsson.se>
%%%----------------------------------------------------------------------

-module(fprof_server).
-author('raimo@erix.ericsson.se').

-behaviour(gen_server).

%% External exports
-export([start/0, call/1, stop/0, stop/1]).
-export([start_trace/0, start_trace/1, start_trace/2, stop_trace/0,
	 spawn/1, spawn/2, spawn/3, spawn/4]).

%% gen_server callbacks
-export([init/1, handle_call/3, handle_cast/2, handle_info/2, 
	 terminate/2, code_change/3]).

-define(TABLE_NAME, fprof_analysis).
-define(DEFAULT_FILESPEC, "fprof.trace").

-include("fprof_server.hrl").

-record(sr, {state = idle,
	     trace_port,
	     trace_pid,
	     fd,
	     table}).

-record(clocks, {
	  id,
	  cnt = 0,   % Number of calls
	  own = 0,   % Own time (wall clock)
	  acc = 0}). % Accumulated time : own + subfunctions (wall clock)

%%%----------------------------------------------------------------------
%%% To be moved to fprof.erl
%%%----------------------------------------------------------------------

%% -export
start_trace() ->
    start_trace_1([], #start_trace{}).

%% -export
start_trace(Args) when list(Args) ->
    start_trace_1(Args, #start_trace{});
start_trace(Arg) ->
    start_trace_1([Arg], #start_trace{}).

%% -export
start_trace(Arg, Value) ->
    start_trace_1([{Arg, Value}], #start_trace{}).



start_trace_1([], #start_trace{
		proc = Proc,
		type = Type,
		mode = Mode,
		file_spec = FileSpec} = StartTrace) ->
    start_trace_2(StartTrace#start_trace{
		    proc      = default(Proc,     self()),
		    type      = default(Type,     file),
		    mode      = default(Mode,     fprof),
		    file_spec = default(FileSpec, ?DEFAULT_FILESPEC)});
start_trace_1([file | Args], 
	      #start_trace{type = undefined} = StartTrace) ->
    start_trace_1([{file, []} | Args], 
		  StartTrace);
start_trace_1([{file, []} | Args], 
	      #start_trace{type = undefined} = StartTrace) ->
    start_trace_1(Args, 
		  StartTrace#start_trace{type = file});
start_trace_1([{file, FileSpec} | Args], 
	      #start_trace{type = undefined} = StartTrace) ->
    start_trace_1(Args, 
		  StartTrace#start_trace{type = file, 
					 file_spec = FileSpec});
start_trace_1([analyse | Args], 
	      #start_trace{type = undefined} = StartTrace) ->
    start_trace_1([{analyse, []} | Args], 
		  StartTrace);
start_trace_1([{analyse, []} | Args], 
	      #start_trace{type = undefined} = StartTrace) ->
    start_trace_1(Args, 
		  StartTrace#start_trace{type = analyse});
start_trace_1([{analyse, FileSpec} | Args], 
	      #start_trace{type = undefined} = StartTrace) ->
    start_trace_1(Args, 
		  StartTrace#start_trace{type = analyse, 
					 file_spec = FileSpec});
start_trace_1([trace | Args], 
	      #start_trace{mode = undefined} = StartTrace) ->
    start_trace_1(Args,
		  StartTrace#start_trace{mode = trace});
start_trace_1([{proc, Proc} | Args], 
	      #start_trace{proc = undefined} = StartTrace) ->
    start_trace_1(Args,
		  StartTrace#start_trace{proc = Proc});
start_trace_1([Arg | _Args], _StartTrace) ->
    erlang:fault(badarg, [Arg]).

start_trace_2(StartTrace) ->
    case fprof_server:start() of
	{ok, _Pid} ->
	    fprof_server:call(StartTrace);
	{error, {already_started, _Pid}} ->
	    fprof_server:call(StartTrace);
	Error ->
	    Error
    end.



stop_trace() ->
    fprof_server:call(#stop_trace{}).



spawn({M, F, Args}) ->
    spawn_1(spawn, M, F, Args, []).

spawn({M, F, Args}, Flags) ->
    spawn_1(spawn, M, F, Args, Flags).

spawn(M, F, Args) ->
    spawn_1(spawn, M, F, Args, []).

spawn(M, F, Args, Flags) ->
    spawn_1(spawn, M, F, Args, Flags).

spawn_1(Spawn, M, F, Args, Flags) 
  when atom(M), atom(F), list(Args), list(Flags) ->
    Parent = self(),
    Ref = make_ref(),
    Pid = erlang:Spawn(
	    fun() ->
		    Parent ! {self(), Ref, ready},
		    MRef = erlang:monitor(process, Parent),
		    receive
			{Parent, Ref, go} ->
			    erlang:demonitor(MRef),
			    receive 
				{'DOWN', MRef, _, _, _} -> ok 
			    after 0 -> ok
			    end;
			{'DOWN', MRef, _, _, _} ->
			    ok
		    end,
		    erlang:apply(M, F, Args)
	    end),
    MRef = erlang:monitor(process, Pid),
    receive 
	{Pid, Ref, ready} -> 
	    erlang:demonitor(MRef),
	    receive 
		{'DOWN', MRef, _, _, Reason} ->
		    exit(Reason)
	    after 0 ->
		    start_trace([{proc, Pid} | Flags]),
		    Pid ! {self(), Ref, go},
		    Pid
	    end;
	{'DOWN', MRef, _, _, Reason} ->
	    exit(Reason)
    end.
	   
%%%----------------------------------------------------------------------
%%% API
%%%----------------------------------------------------------------------
start() ->
    gen_server:start({local, ?FPROF_SERVER}, ?MODULE, [], []).



call(Request) ->
    io:format("~p:call(~p)~n", [?MODULE, Request]),
    gen_server:call(?FPROF_SERVER, Request, infinity).



stop() ->
    stop(normal).

stop(Reason) ->
    call(#stop{reason = Reason}).

%%%----------------------------------------------------------------------
%%% Callback functions from gen_server
%%%----------------------------------------------------------------------

%%----------------------------------------------------------------------
%% Func: init/1
%% Returns: {ok, State}          |
%%          {ok, State, Timeout} |
%%          ignore               |
%%          {stop, Reason}
%%----------------------------------------------------------------------
init([]) ->
    process_flag(trap_exit, true),
    {ok, #sr{}}.

%%----------------------------------------------------------------------
%% Func: handle_call/3
%% Returns: {reply, Reply, State}          |
%%          {reply, Reply, State, Timeout} |
%%          {noreply, State}               |
%%          {noreply, State, Timeout}      |
%%          {stop, Reason, Reply, State}   | (terminate/2 is called)
%%          {stop, Reason, State}            (terminate/2 is called)
%%----------------------------------------------------------------------
handle_call(#start_trace{proc = Proc,
			 type = file,
			 mode = Mode, 
			 file_spec = FileSpec},
	    From, #sr{trace_port = undefined,
		      trace_pid = undefined} = SR) ->
    trace_off(),
    Port = mk_trace_port(FileSpec),
    trace_on(Proc, Port, Mode),
    {reply, ok, SR#sr{trace_port = Port}};
handle_call(#start_trace{proc = Proc,
			 type = analyse,
			 mode = Mode,
			 file_spec = FileSpec},
	    From, #sr{trace_port = undefined,
		      trace_pid = undefined} = SR) ->
    trace_off(),
    {ok, FD} = file:open(FileSpec, [write]),
    Pid = mk_trace_pid(FD),
    trace_on(Proc, Pid, Mode),
    {reply, {ok, Pid}, SR#sr{trace_pid = Pid, fd = FD}};
handle_call(#start_trace{}, 
	    From, SR) ->
    {reply, {error, {already_tracing}}, SR};
handle_call(#stop_trace{}, 
	    From, #sr{trace_port = undefined,
		      trace_pid = undefined} = SR) ->
    {reply, {error, {not_tracing}}, SR};
handle_call(#stop_trace{}, 
	    From, #sr{trace_port = Port, 
		      trace_pid = undefined} = SR) ->
    trace_off(),
    catch erlang:port_close(Port),
    {reply, ok, SR#sr{trace_port = undefined}};
handle_call(#stop_trace{}, 
	    From, #sr{trace_port = undefined,
		      trace_pid = Pid, 
		      fd = FD} = SR) ->
    trace_off(),
    exit(Pid, stop),
    file:close(FD),
    {reply, ok, SR#sr{trace_pid = undefined}};
handle_call(#stop{reason = Reason}, From, SR) ->
    {stop, Reason, ok, SR};
handle_call(state, From, SR) ->
    {reply, {ok, SR}, SR};
handle_call(Request, From, State) ->
    error_logger:info_msg("~p:handle_call, unknown request - ~p", [Request]),
    {noreply, State}.

%%----------------------------------------------------------------------
%% Func: handle_cast/2
%% Returns: {noreply, State}          |
%%          {noreply, State, Timeout} |
%%          {stop, Reason, State}            (terminate/2 is called)
%%----------------------------------------------------------------------
handle_cast(Msg, State) ->
    error_logger:info_msg("~p:handle_cast, unknown msg - ~p", [Msg]),
    {noreply, State}.

%%----------------------------------------------------------------------
%% Func: handle_info/2
%% Returns: {noreply, State}          |
%%          {noreply, State, Timeout} |
%%          {stop, Reason, State}            (terminate/2 is called)
%%----------------------------------------------------------------------
handle_info({'EXIT', Id, Reason} = Info, 
	    #sr{trace_port = TracePort,
		trace_pid = TracePid} = SR) ->
    case Id of
	TracePort ->
	    error_logger:info_msg(
	      "~p:handle_info, unexpected ~p (trace port)", 
	      [?MODULE, Info]);
	TracePid ->
	    error_logger:info_msg(
	      "~p:handle_info, unexpected ~p (trace pid)", 
	      [?MODULE, Info]);
	_ ->
	    ok
    end,
    {noreply, SR};
handle_info(Info, SR) ->
    error_logger:info_msg("~p:handle_info, unknown info - ~p", 
			  [?MODULE, Info]),
    {noreply, SR}.

%%----------------------------------------------------------------------
%% Func: terminate/2
%% Purpose: Shutdown the server
%% Returns: any (ignored by gen_server)
%%----------------------------------------------------------------------
terminate(Reason, State) ->
    ok.

%%----------------------------------------------------------------------
%% Func: code_change/3
%% Purpose: Convert process state when code is changed
%% Returns: {ok, NewState}
%%----------------------------------------------------------------------
code_change(OldVsn, State, Extra) ->
    {ok, State}.

%%%----------------------------------------------------------------------
%%% Internal functions
%%%----------------------------------------------------------------------

mk_trace_pid(Dest) ->
    spawn_link(
      fun() ->
	      trace_pid(Dest)
      end).

trace_pid(Dest) ->
    receive
	Term ->
	    io:format(Dest, "~p~n", [Term]),
	    trace_pid(Dest)
    end.

mk_trace_port(FileSpec) ->
    TracePortFun = dbg:trace_port(file, FileSpec),
    TracePortFun().

trace_off() ->
    erlang:trace(all, false,
		 [send, 'receive', procs, call, silent, return_to,
		  running, garbage_collection, timestamp, arity, 
		  set_on_spawn, set_on_first_spawn, 
		  set_on_link, set_on_first_link]),
    erlang:trace_pattern({'_', '_', '_'}, false, []),
    erlang:trace_pattern(on_load, false, []).

trace_on(Proc, Tracer, fprof) ->    
    erlang:trace_pattern({'_', '_', '_'}, true, []),
    erlang:trace_pattern(on_load, true, []),
    my_trace(Proc, true, 
	     [call, return_to, 
	      running, procs, garbage_collection, 
	      arity, timestamp, set_on_spawn,
	      {tracer, Tracer}]);
trace_on(Proc, Tracer, trace) ->
    erlang:trace_pattern({'_', '_', '_'}, true, []),
    erlang:trace_pattern(on_load, true, []),
    my_trace(Proc, true, 
	     [call, return_to, send, 'receive', procs,
	      set_on_spawn, {tracer, Tracer}]).

my_trace([], _How, _FlagList) ->
    ok;
my_trace([PidSpec | Tail], How, FlagList) ->
    erlang:trace(PidSpec, How, FlagList),
    my_trace(Tail, How, FlagList);
my_trace(PidSpec, How, FlagList) ->
    erlang:trace(PidSpec, How, FlagList),
    ok.

default(undefined, Value) ->
    Value;
default(Option, Value) ->
    Option.
